// BI1MMXBase.cpp: implementation of the CBI1MMXBase class.
//
//////////////////////////////////////////////////////////////////////

//#include "stdafx.h"
#include "BI1MMXBase.h"
#include <memory.h>
#include <stdio.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#ifndef _MSC_VER

#define rotr32(x,n)   (((x) >> ((int)(n))) | ((x) << (32 - (int)(n))))
#define rotl32(x,n)   (((x) << ((int)(n))) | ((x) >> (32 - (int)(n))))

#else

#include <stdlib.h>

#pragma intrinsic(_lrotr,_lrotl)
#define rotr32(x,n)   _lrotr(x,n)
#define rotl32(x,n)   _lrotl(x,n)

#endif

inline void rotl64(u4byte& x, u4byte& y, u4byte& n){
        u4byte tmp;
        if(n==0) return;
        if((unsigned int) n < 32){
            tmp = x;
            x = ((x) << ((int) n)) | ((y) >> (32 - (int)(n)));  
            y = ((y) << ((int) n)) | ((tmp) >> (32 - (int)(n)));  
        }else if(n==32){
            tmp=x;
            x=y;
            y=tmp;
        }else{
            n -= 32;
            tmp = y;
            y = ((x) << ((int) n)) | ((y) >> (32 - (int)(n)));  
            x = ((tmp) << ((int) n)) | ((x) >> (32 - (int)(n)));  
        }                                                       
}

inline void rotr64(u4byte& x, u4byte& y, u4byte& n){
        u4byte tmp;
        if(n==0) return;
        if((unsigned int) n < 32){
            tmp=x;
            x = ((x) >> ((int) n)) | ((y) << (32 - (int)(n)));  
            y = ((y) >> ((int) n)) | ((tmp) << (32 - (int)(n)));  
        }else if(n==32){
            tmp=x;
            x=y;
            y=tmp;
        }else{
            n -= 32;                      
            tmp=y;
            y = ((x) >> ((int) n)) | ((y) << (32 - (int)(n)));  
            x = ((tmp) >> ((int) n)) | ((x) << (32 - (int)(n)));  
        }                                                       
}


void addbyte(u4byte& a,u4byte& b,u4byte& c,u4byte& d){
    BYTE *x,*y;
    x = (BYTE*) &a;
    y = (BYTE*) &c;
    x[0]+=y[0]; x[1]+=y[1]; x[2]+=y[2]; x[3]+=y[3];
    x = (BYTE*) &b;
    y = (BYTE*) &d;
    x[0]+=y[0]; x[1]+=y[1]; x[2]+=y[2]; x[3]+=y[3];
}


void subbyte(u4byte& a,u4byte& b,u4byte& c,u4byte& d){
    BYTE *x,*y;
    x = (BYTE*) &a;
    y = (BYTE*) &c;
    x[0]-=y[0]; x[1]-=y[1]; x[2]-=y[2]; x[3]-=y[3];
    x = (BYTE*) &b;
    y = (BYTE*) &d;
    x[0]-=y[0]; x[1]-=y[1]; x[2]-=y[2]; x[3]-=y[3];
}


/* MMX instruction */
int CheckBI1MMX(void){
    int MMX_Flag;
#define cpuid __asm _emit 0x0f __asm _emit 0xa2
    __asm {
        mov MMX_Flag, 0
        pushfd
        pop eax
        mov ecx, eax
        xor eax, 0x200000
        push eax
        popfd
        pushfd
        pop eax
        xor eax, ecx
        jz NO_MMX_TECH
        mov eax, 1
        cpuid
        test edx, 0x800000
        jz NO_MMX_TECH
        mov MMX_Flag, 1
NO_MMX_TECH:
    }
    return MMX_Flag;
}


/* initialise the key schedule from the user supplied key   */


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CBI1MMXBase::CBI1MMXBase()
{
	memset(m_key,0,44 * 4);
    if(CheckBI1MMX()) IsMMX=1;
    else IsMMX=0;
    m_KeySize=16;
}

CBI1MMXBase::~CBI1MMXBase()
{

}

DWORD CBI1MMXBase::DWORD_Sub(DWORD A, DWORD B)
{
	if(A >= B) return A - B;
	
	B = 0xffffffff - B + 1;
	return A + B;
}


// virtual functions

int CBI1MMXBase::CipherInitial(DWORD dwParam)
{
	return 0;
}

int CBI1MMXBase::KeyInitial(BYTE *pByte, DWORD nKeySize, DWORD dwParam)
{
	if(nKeySize > 32) return -1;

	m_KeySize = nKeySize;

	DWORD dwKeySize = nKeySize / 4;
	DWORD *pDWORD = (DWORD *)pByte;

	u4byte  i, j, k, a, b, l[8];

    m_key[0] = 0xb7e15163;

    for(k = 1; k < 44; ++k)
        
        m_key[k] = m_key[k - 1] + 0x9e3779b9;

    for(k = 0; k < dwKeySize; ++k)

        l[k] = pDWORD[k];

    for(a = b = k = 0; k < 132; ++k)
    {
        i = k % 44; 
		j = k % dwKeySize;
        a = rotl32(m_key[i] + a + b, 3); b += a;
        b = rotl32(l[j] + b, b & 0x0000001f);
        m_key[i] = a; l[j] = b;
    }
	return 0;
}

int CBI1MMXBase::Encrypt(BYTE *pByte, DWORD nBlockSize)
{
    if(!IsMMX){
	    u4byte a,b,c,d,t;

    	DWORD *pDWORD = (DWORD *)pByte;
    
        a = pDWORD[0]; b = pDWORD[1];
        c = pDWORD[2]; d = pDWORD[3];

        addbyte(a,b,m_key[0], m_key[1]);
        addbyte(c,d,m_key[2], m_key[3]);

        for(int i=4; i<44; i+=4){
            //printf("round i=%i a=%x b=%x c=%x d=%x\n",i,a,b,c,d); 
            t= c*d >> 26;
            //printf("  t=%x ", t);
            a ^= c; b ^= d;
            //printf("a^c=%x b^c=%x\n", a,b);
            rotl64(a,b,t);
            //printf("  after rotl64 a=%x b=%x\n", a,b);
            addbyte(a,b,m_key[i], m_key[i+1]);
            //printf("  final a=%x b=%x\n", a,b);
            t= a*b >> 26;
            //printf("  t=%x ", t);
            c ^= a; d ^= b;
            //printf("c^a=%x d^b=%x\n", c,d);
            rotl64(c,d,t);
            //printf("  after rotl64 c=%x d=%x\n", c,d);
            addbyte(c,d,m_key[i+2], m_key[i+3]);
            //printf("  final c=%x d=%x\n", c,d);
            //printf("  a=%x b=%x c=%x d=%x\n", a,b,c,d);
        }
        pDWORD[0] = a; pDWORD[1] = b;
        pDWORD[2] = c; pDWORD[3] = d;
        return 0;
    }
    unsigned long *pkey;
    pkey=m_key;

    __asm{
        mov esi, pByte
        movq mm1, [esi]    //mm1: B A
        movq mm2, [esi+8]  //mm2: D C

        mov esi, pkey

        movq mm0, [esi]    //mm0: S[1] S[0]
        paddb mm1, mm0     //(B,A) += (S[1],S[0])
        
        add esi, 8
        movq mm0, [esi]    //mm0: S[3] S[2]
        paddb mm2, mm0     //(D,C) += (S[3],S[2])


        mov ecx, 10     // 10 rounds

BeginRound:

        movd eax, mm2   //eax: C
        movq mm3, mm2   //mm3: D C
        psrlq mm3, 32   //mm3: 0 D
        movd ebx, mm3   //ebx: D
        mul ebx
        shr eax, 26     //eax=(C*D)>>26
        movd mm4, eax   //mm4=eax
        mov ebx, 64
        sub ebx, eax
        movd mm5, ebx   //mm5=64-eax
        pxor mm1, mm2
        movq mm6, mm1
        psllq mm1, mm4
        psrlq mm6, mm5
        por mm1, mm6
        add esi, 8
        movq mm0, [esi]
        paddb mm1, mm0

        movd eax, mm1   //eax: A
        movq mm3, mm1   //mm3: B A
        psrlq mm3, 32   //mm3: 0 B
        movd ebx, mm3   //ebx: B
        mul ebx
        shr eax, 26     //eax=(A*B)>>26
        movd mm4, eax   //mm4=eax
        mov ebx, 64
        sub ebx, eax
        movd mm5, ebx   //mm5=64-eax
        pxor mm2, mm1
        movq mm6, mm2
        psllq mm2, mm4
        psrlq mm6, mm5
        por mm2, mm6
        add esi, 8
        movq mm0, [esi]
        paddb mm2, mm0

        loop BeginRound// 10 times

        mov esi, pByte
        movq [esi], mm1
        movq [esi+8], mm2
        emms
    }
    return 0;
}

int CBI1MMXBase::Decrypt(BYTE *pByte, DWORD nBlockSize)
{
    if(!IsMMX){
	    u4byte  a,b,c,d,t;
	    DWORD *pDWORD = (DWORD *)pByte;

        a = pDWORD[0]; b = pDWORD[1];
        c = pDWORD[2]; d = pDWORD[3];

        for(int i=40; i>0; i-=4){
            t=a*b >> 26;
            subbyte(c,d,m_key[i+2], m_key[i+3]);
            rotr64(c,d, t);
            c^=a; d^=b;
            t=c*d >> 26;
            subbyte(a,b,m_key[i], m_key[i+1]);
            rotr64(a,b, t);
            a^=c; b^=d;
        }

        subbyte(c,d,m_key[2], m_key[3]);
        subbyte(a,b,m_key[0], m_key[1]);

        pDWORD[3] = d; pDWORD[2] = c; 
        pDWORD[1] = b; pDWORD[0] = a; 
	    return 0;
    }
    unsigned long *pkey;
    pkey=&(m_key[42]);

    __asm{
        mov esi, pByte
        movq mm1, [esi]    //mm1: B A
        movq mm2, [esi+8]  //mm2: D C
        mov esi, pkey
        mov ecx, 10     // 10 rounds

BeginRound:

        movd eax, mm1   //eax: A
        movq mm3, mm1   //mm3: B A
        psrlq mm3, 32   //mm3: 0 B
        movd ebx, mm3   //ebx: B
        mul ebx
        shr eax, 26     //eax=(A*B)>>26
        movd mm4, eax   //mm4=eax
        mov ebx, 64
        sub ebx, eax
        movd mm5, ebx   //mm5=64-eax
        movq mm0, [esi]
        psubb mm2, mm0
        sub esi, 8
        movq mm6, mm2
        psrlq mm2, mm4
        psllq mm6, mm5
        por mm2, mm6
        pxor mm2, mm1

        movd eax, mm2   //eax: C
        movq mm3, mm2   //mm3: D C
        psrlq mm3, 32   //mm3: 0 D
        movd ebx, mm3   //ebx: D
        mul ebx
        shr eax, 26     //eax=(C*D)>>26
        movd mm4, eax   //mm4=eax
        mov ebx, 64
        sub ebx, eax
        movd mm5, ebx   //mm5=64-eax
        movq mm0, [esi]
        psubb mm1, mm0
        sub esi, 8
        movq mm6, mm1
        psrlq mm1, mm4
        psllq mm6, mm5
        por mm1, mm6
        pxor mm1, mm2

        loop BeginRound// 10 times

        movq mm0, [esi]    //mm0: S[3] S[2]
        psubb mm2, mm0     //(D,C) -= (S[3],S[2])
        sub esi, 8

        movq mm0, [esi]    //mm0: S[1] S[0]
        psubb mm1, mm0     //(B,A) -= (S[1],S[0])
        
        mov esi, pByte
        movq [esi], mm1
        movq [esi+8], mm2
        emms
    }
    return 0;
}

int CBI1MMXBase::GetBlockSize()
{
	return 16;
}

int CBI1MMXBase::GetKeySize() 
{
	return m_KeySize;
}

int CBI1MMXBase::GetRounds()
{
	return 16;
}

