// RC6MMXBase.cpp: implementation of the CRC6MMXBase class.
//
//////////////////////////////////////////////////////////////////////

//#include "stdafx.h"
#include "RC6MMXBase.h"
#include <memory.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#ifndef _MSC_VER

#define rotr(x,n)   (((x) >> ((int)(n))) | ((x) << (32 - (int)(n))))
#define rotl(x,n)   (((x) << ((int)(n))) | ((x) >> (32 - (int)(n))))

#else

#include <stdlib.h>

#pragma intrinsic(_lrotr,_lrotl)
#define rotr(x,n)   _lrotr(x,n)
#define rotl(x,n)   _lrotl(x,n)

#endif

#define f_rnd(i,a,b,c,d)                                \
        t = rotl(b * (b + b + 1), 5);                   \
        u = rotl(d * (d + d + 1), 5);                   \
        a = rotl(a ^ t, u & 0x0000001f) + m_key[i];     \
        c = rotl(c ^ u, t & 0x0000001f) + m_key[i + 1]  \

#define i_rnd(i,a,b,c,d)                                \
        u = rotl(d * (d + d + 1), 5);                   \
        t = rotl(b * (b + b + 1), 5);                   \
        c = rotr(DWORD_Sub(c, m_key[i + 1]), t & 0x0000001f) ^ u; \
        a = rotr(DWORD_Sub(a, m_key[i]), u & 0x0000001f) ^ t      \

/* MMX instruction */
int CheckMMX(void){
    int MMX_Flag;
#define cpuid __asm _emit 0x0f __asm _emit 0xa2
    __asm {
        mov MMX_Flag, 0
        pushfd
        pop eax
        mov ecx, eax
        xor eax, 0x200000
        push eax
        popfd
        pushfd
        pop eax
        xor eax, ecx
        jz NO_MMX_TECH
        mov eax, 1
        cpuid
        test edx, 0x800000
        jz NO_MMX_TECH
        mov MMX_Flag, 1
NO_MMX_TECH:
    }
    return MMX_Flag;
}


/* initialise the key schedule from the user supplied key   */


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRC6MMXBase::CRC6MMXBase()
{
	memset(m_key,0,44 * 4);
    if(CheckMMX()) IsMMX=1;
    else IsMMX=0;
}

CRC6MMXBase::~CRC6MMXBase()
{

}

DWORD CRC6MMXBase::DWORD_Sub(DWORD A, DWORD B)
{
	if(A >= B) return A - B;
	
	B = 0xffffffff - B + 1;
	return A + B;
}


// virtual functions

int CRC6MMXBase::CipherInitial(DWORD dwParam)
{
	return 0;
}

int CRC6MMXBase::KeyInitial(BYTE *pByte, DWORD nKeySize, DWORD dwParam)
{
	if(nKeySize > 32) return -1;

	m_KeySize = nKeySize;

	DWORD dwKeySize = nKeySize / 4;
	DWORD *pDWORD = (DWORD *)pByte;

	u4byte  i, j, k, a, b, l[8];

    m_key[0] = 0xb7e15163;

    for(k = 1; k < 44; ++k)
        
        m_key[k] = m_key[k - 1] + 0x9e3779b9;

    for(k = 0; k < dwKeySize; ++k)

        l[k] = pDWORD[k];

    for(a = b = k = 0; k < 132; ++k)
    {
        i = k % 44; 
		j = k % dwKeySize;
        a = rotl(m_key[i] + a + b, 3); b += a;
        b = rotl(l[j] + b, b & 0x0000001f);
        m_key[i] = a; l[j] = b;
    }
	return 0;
}

int CRC6MMXBase::Encrypt(BYTE *pByte, DWORD nBlockSize)
{
    if(!IsMMX){
	    u4byte  a,b,c,d,t,u;

    	DWORD *pDWORD = (DWORD *)pByte;
    
        a = pDWORD[0]; b = pDWORD[1] + m_key[0];
        c = pDWORD[2]; d = pDWORD[3] + m_key[1];
    
        f_rnd( 2,a,b,c,d); f_rnd( 4,b,c,d,a);
        f_rnd( 6,c,d,a,b); f_rnd( 8,d,a,b,c);
        f_rnd(10,a,b,c,d); f_rnd(12,b,c,d,a);
        f_rnd(14,c,d,a,b); f_rnd(16,d,a,b,c);
        f_rnd(18,a,b,c,d); f_rnd(20,b,c,d,a);
        f_rnd(22,c,d,a,b); f_rnd(24,d,a,b,c);
        f_rnd(26,a,b,c,d); f_rnd(28,b,c,d,a);
        f_rnd(30,c,d,a,b); f_rnd(32,d,a,b,c);
        f_rnd(34,a,b,c,d); f_rnd(36,b,c,d,a);
        f_rnd(38,c,d,a,b); f_rnd(40,d,a,b,c);

        pDWORD[0] = a + m_key[42]; pDWORD[1] = b;
        pDWORD[2] = c + m_key[43]; pDWORD[3] = d;
	    return 0;
    }
    unsigned long *pkey;
    pkey=m_key;
    __asm{
        mov esi, pByte
        movd mm1, [esi]
        movd mm2, [esi+4]
        movd mm4, [esi+8]
        movd mm3, [esi+12]
        psllq mm4, 32
        por mm1, mm4  //mm1: C A, mm2: 0 B, mm3: 0 D

        mov esi, pkey

        movd mm0, [esi]
        paddd mm2, mm0  // B+S[0]
        movd mm0, [esi+4]
        paddd mm3, mm0
        mov ecx, 20     // 20 rounds

BeginRound:

        movd eax, mm2 //B
        mov ebx, eax
        add eax, ebx
        inc eax
        mul ebx
        rol eax, 5
        movd mm7, eax //t
        movd eax, mm3 //D
        mov ebx, eax
        add eax, ebx
        inc eax
        mul ebx
        rol eax, 5
        movd mm4, eax //u
        psllq mm4, 32
        por mm4, mm7 // mm4: u t

        pxor mm1, mm4

        mov ebx, ecx

        and eax, 0x1F //u&0x1F
        mov cl, al
        movd eax, mm1
        rol eax, cl
        movd mm7, eax // mm7: 0 A

        psrlq mm1, 32
        movd eax, mm4 //t
        
        and eax, 0x1F //t&0x1F
        mov cl, al
        movd eax, mm1
        rol eax, cl
        movd mm1, eax // mm1: 0 C

        psllq mm1, 32
        por mm1, mm7 // mm1: C A

        mov ecx, ebx

        add esi, 8
        movq mm0, [esi]
        paddd mm1, mm0

        movq mm7, mm1 // permutation
        movq mm1, mm3
        psllq mm1, 32
        por mm1, mm2
        movq mm2, mm7
        psrlq mm2, 32
        movd eax, mm7
        movd mm3, eax

        loop BeginRound// 20 times

        add esi, 8
        movq mm0, [esi]
        paddd mm1, mm0

        mov esi, pByte
        movd [esi], mm1
        movd [esi+4], mm2
        psrlq mm1, 32
        movd [esi+8], mm1
        movd [esi+12], mm3
        emms
    }
    return 0;
}

int CRC6MMXBase::Decrypt(BYTE *pByte, DWORD nBlockSize)
{
	u4byte  a,b,c,d,t,u;
	DWORD *pDWORD = (DWORD *)pByte;

    d = pDWORD[3]; c = DWORD_Sub(pDWORD[2] , m_key[43]); 
    b = pDWORD[1]; a = DWORD_Sub(pDWORD[0] , m_key[42]);

    i_rnd(40,d,a,b,c); i_rnd(38,c,d,a,b);
    i_rnd(36,b,c,d,a); i_rnd(34,a,b,c,d);
    i_rnd(32,d,a,b,c); i_rnd(30,c,d,a,b);
    i_rnd(28,b,c,d,a); i_rnd(26,a,b,c,d);
    i_rnd(24,d,a,b,c); i_rnd(22,c,d,a,b);
    i_rnd(20,b,c,d,a); i_rnd(18,a,b,c,d);
    i_rnd(16,d,a,b,c); i_rnd(14,c,d,a,b);
    i_rnd(12,b,c,d,a); i_rnd(10,a,b,c,d);
    i_rnd( 8,d,a,b,c); i_rnd( 6,c,d,a,b);
    i_rnd( 4,b,c,d,a); i_rnd( 2,a,b,c,d);

    pDWORD[3] = DWORD_Sub(d , m_key[1]); pDWORD[2] = c; 
    pDWORD[1] = DWORD_Sub(b , m_key[0]); pDWORD[0] = a; 
	return 0;
}

int CRC6MMXBase::GetBlockSize()
{
	return 16;
}

int CRC6MMXBase::GetKeySize() 
{
	return m_KeySize;
}

int CRC6MMXBase::GetRounds()
{
	return 16;
}

