// SBase.cpp: implementation of the CSBase class.
//
//////////////////////////////////////////////////////////////////////

//#include "stdafx.h"
#include "SBase.h"
#include <memory.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#ifndef _MSC_VER

#define rotr(x,n)   (((x) >> ((int)(n))) | ((x) << (32 - (int)(n))))
#define rotl(x,n)   (((x) << ((int)(n))) | ((x) >> (32 - (int)(n))))

#else

#include <stdlib.h>

#pragma intrinsic(_lrotr,_lrotl)
#define rotr(x,n)   _lrotr(x,n)
#define rotl(x,n)   _lrotl(x,n)

#endif
inline void rotl64(u4byte& x, u4byte& y, int n){
        u4byte tmp;
        if(n==0) return;
        if((unsigned int) n < 32){
            tmp = x;
            x = ((x) << ((int) n)) | ((y) >> (32 - (int)(n)));  
            y = ((y) << ((int) n)) | ((tmp) >> (32 - (int)(n)));  
        }else if(n==32){
            tmp=x;
            x=y;
            y=tmp;
        }else{
            n -= 32;
            tmp = y;
            y = ((x) << ((int) n)) | ((y) >> (32 - (int)(n)));  
            x = ((tmp) << ((int) n)) | ((x) >> (32 - (int)(n)));  
        }                                                       
}

inline void rotr64(u4byte& x, u4byte& y, int n){
        u4byte tmp;
        if(n==0) return;
        if((unsigned int) n < 32){
            tmp=x;
            x = ((x) >> ((int) n)) | ((y) << (32 - (int)(n)));  
            y = ((y) >> ((int) n)) | ((tmp) << (32 - (int)(n)));  
        }else if(n==32){
            tmp=x;
            x=y;
            y=tmp;
        }else{
            n -= 32;                      
            tmp=y;
            y = ((x) >> ((int) n)) | ((y) << (32 - (int)(n)));  
            x = ((tmp) >> ((int) n)) | ((x) << (32 - (int)(n)));  
        }                                                       
}


/* MMX instruction */
int CheckSMMX(void){
    int MMX_Flag;
#define cpuid __asm _emit 0x0f __asm _emit 0xa2
    __asm {
        mov MMX_Flag, 0
        pushfd
        pop eax
        mov ecx, eax
        xor eax, 0x200000
        push eax
        popfd
        pushfd
        pop eax
        xor eax, ecx
        jz NO_MMX_TECH
        mov eax, 1
        cpuid
        test edx, 0x800000
        jz NO_MMX_TECH
        mov MMX_Flag, 1
NO_MMX_TECH:
    }
    return MMX_Flag;
}


/* initialise the key schedule from the user supplied key   */


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSBase::CSBase()
{
	memset(m_key,0,32 * 4);
    if(CheckSMMX()) IsMMX=1;
    else IsMMX=0;
    m_KeySize=16;
}

CSBase::~CSBase()
{

}


// virtual functions

int CSBase::CipherInitial(DWORD dwParam)
{
	return 0;
}

int CSBase::KeyInitial(BYTE *pByte, DWORD nKeySize, DWORD dwParam)
{
    u4byte Key1X, Key1Y, Key2X, Key2Y;
    Key1X=mgetBLong(pByte); Key1Y=mgetBLong(pByte);
    Key2X=mgetBLong(pByte); Key2Y=mgetBLong(pByte);

    for(int i=0; i<32;){
        rotr64(Key1X,Key1Y,4);
        rotl64(Key2X,Key2Y,4);
        m_key[i]=Key1X^Key2X; i++;
        m_key[i]=Key1Y^Key2Y; i++;
    }

    return 0;
}

int CSBase::Encrypt(BYTE *pByte, DWORD nBlockSize)
{
    if(!IsMMX){
	    u4byte  a,b;

    	DWORD *pDWORD = (DWORD *)pByte;
    
        b = pDWORD[0]; a = pDWORD[1];
        for(int i=0; i<32;){
            a = ((((a>>16) * 207) & 0xFFFF) << 16) | 
                (((a&0xFFFF)*207) & 0xFFFF);
            b = ((((b>>16) * 207) & 0xFFFF) << 16) | 
                (((b&0xFFFF)*207) & 0xFFFF);
            rotr64(a,b,9);
            b ^= m_key[i]; i++;
            a ^= m_key[i]; i++;
        }
        pDWORD[0] = b; pDWORD[1] = a;
	    return 0;
    }
    unsigned long *pkey;
    pkey=m_key;
    unsigned int num[2]={0x00CF00CF, 0x00CF00CF};
    unsigned int *magic;
    magic=num;
    __asm{
        mov esi, magic
        movq mm0, [esi]
        mov esi, pByte
        movq mm1, [esi] //mm1: a b
        mov edi, pkey

        mov ecx, 16     // 16 rounds

BeginRound:

        pmullw mm1, mm0
        movq mm3, mm1
        psrlq mm1, 9
        psllq mm3, 55
        por mm1, mm3
        pxor mm1, [edi]
        add edi, 8

        loop BeginRound// 20 times

        movq [esi], mm1
        emms
    }
    return 0;
}


int CSBase::Decrypt(BYTE *pByte, DWORD nBlockSize)
{
    if(!IsMMX){
	    u4byte  a,b;
	    DWORD *pDWORD = (DWORD *)pByte;
    
        b = pDWORD[0]; a = pDWORD[1];
        for(int i=31; i>=0;){
            a ^= m_key[i]; i--;
            b ^= m_key[i]; i--;
            rotl64(a,b,9);
            a = ((((a>>16) * 1583) & 0xFFFF) << 16) | 
                (((a&0xFFFF)*1583) & 0xFFFF);
            b = ((((b>>16) * 1583) & 0xFFFF) << 16) | 
                (((b&0xFFFF)*1583) & 0xFFFF);
        }
        pDWORD[0] = b; pDWORD[1] = a; 
	    return 0;
    }

    unsigned long *pkey;
    pkey=m_key;
    unsigned int num[2]={ 0x062F062F, 0x062F062F};
    unsigned int* magic;
    magic=num;
    __asm{
        mov esi, magic
        movq mm0, [esi]
        mov esi, pByte
        movq mm1, [esi] //mm1: a b
        mov edi, pkey
        add edi, 120

        mov ecx, 16     // 16 rounds

BeginRound:

        pxor mm1, [edi]
        sub edi, 8
        movq mm3, mm1
        psllq mm1, 9
        psrlq mm3, 55
        por mm1, mm3
        pmullw mm1, mm0

        loop BeginRound// 20 times

        movq [esi], mm1
        emms
    }
    return 0;
}

int CSBase::GetBlockSize()
{
	return 8;
}

int CSBase::GetKeySize() 
{
	return m_KeySize;
}

int CSBase::GetRounds()
{
	return 16;
}

