#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "CipherBase.h"
#include "xortab.h"
#include "SBase.h"
#include "BI2MMXBase.h"

extern void genData(BYTE*);
extern void genKey(BYTE*);
extern void showBits(BYTE*, int);

// n is number of bits
//  xor the first n bits of data
inline int xor128(BYTE* data, int n){
    int r;
    r = n%16;
    n /= 16;
    unsigned short temp;
    int result=0;
    for(int i=0; i<n; i++){
        temp = mgetBWord(data);
        result ^= xor16[temp];
    }
    if(r!=0){
        temp = mgetBWord(data);
        temp >>= (16-r);
        result ^= xor16[temp];
    }
    return result;
}

inline int sepXor128(BYTE* data, int* dataBit, int dataNum){
    int result=0;
    for(int i=0; i<dataNum; i++){
        int r=dataBit[i]%8;
        int q=dataBit[i]/8;
        result ^= (data[q] >> (7-r)) & 0x01;
    }
    return result;
}

/* testing function for xor128(..) 
int xor128test(BYTE* data, int n){
    int r=n%8;
    int q=n/8;
    int result=0;
    for(int i=0; i<q; i++){
        unsigned int temp = (unsigned int) data[i];
        for(int j=0; j<8; j++){
            result ^= (temp>>j) & 0x01;
        }
    }
    if(r!=0){
        unsigned int temp = (unsigned int) data[i];
        temp >>= (8-r);
        for(int j=0; j<r; j++) result ^= (temp>>j) & 0x01;
    }
    return result;
}

void testxor128(void){
    BYTE data[16];
    for(int i=0; i<10000; i++){
        genData(data);
        //BYTE* t=data;
        //for(int k=0; k<8; k++){
        //    showBits(t,1);printf(" "); t++; showBits(t, 1);t++; printf("\n");
        //}
        int dataBit[128];
        for(int j=0; j<128; j++){
            for(int k=0; k<j; k++) dataBit[k]=k;
            if(xor128(data,j) != sepXor128(data, dataBit, j)){
                printf("bug!!!\n");
                printf("j=%d, xor128=%d, xor128test=%d\n", j, xor128(data,j), sepXor128(data, dataBit,j));
            }
        }
    }
}
*/

// run n times
void linearAtt(CCipherBase* c, int n, char* filename){
    static int stat[128][128][128];
    //   key   in  out
    for(int i=0; i<128; i++)
        for(int j=0; j<128; j++)
            for(int k=0; k<128; k++)
                stat[i][j][k]=0;

    FILE* fp;
    fp=fopen(filename, "w");
    clock_t begin, end;
    begin = clock();

    c->CipherInitial(0);
    int blocksize=c->GetBlockSize();
    int keysize;
    BYTE input[16], output[16], key[16];
    for(int a=0; a<n; a++){
        if(a%10 ==0){
            end=clock();
            printf("Time=%d(msec) times=%d\n", end-begin, a);
        }
        genData(input);
        for(int b=0; b<16; b++) output[b]=input[b];
        genKey(key);
        c->KeyInitial(key,16,0);
        c->Encrypt(output, blocksize);
        keysize=c->GetKeySize();
        for(int i=1; i<=keysize*8; i++)
            for(int j=1; j<=blocksize*8; j++)
                for(int k=1; k<=blocksize*8; k++)
                    stat[i-1][j-1][k-1] += xor128(input,j) ^ xor128(output,k) ^ xor128(key,i);

    }
    for(i=0; i<keysize*8; i++){
        int min=stat[i][0][0];
        int max=stat[i][0][0];
        int minJ=0, minK=0, maxJ=0, maxK=0;
        for(int j=0; j<blocksize*8; j++){
            for(int k=0; k<blocksize*8; k++){
                if(stat[i][j][k] > max){
                    max = stat[i][j][k];
                    maxJ=j; maxK=k;
                }
                if(stat[i][j][k] < min){
                    min = stat[i][j][k];
                    minJ=j; minK=k;
                }
            }
        }
        fprintf(fp, "MAX: Key[1..%d] = P[1..%d] xor C[1..%d] with probability %f\n", i+1, maxJ+1, maxK+1, 1-((double) max)/(n*1.0));
        fprintf(fp, "MIN: Key[1..%d] = P[1..%d] xor C[1..%d] with probability %f\n\n", i+1, minJ+1, minK+1, 1-((double) min)/(n*1.0));
    }
    printf("Finished!! check linear.txt\n");
}



void linearAtt2(CCipherBase* c, int n1, double t1, int n2, double t2, int sec){
    FILE* fp;
    fp=fopen("linear2.txt", "w");
    clock_t begin, end;
    begin = clock();

    c->CipherInitial(0);
    int blocksize=c->GetBlockSize();
    int keysize=c->GetKeySize();
    int blockbits=blocksize*8;
    int keybits=keysize*8;
    BYTE input[16], output[16], key[16];
    int inBit[128], outBit[128], keyBit[128];
    int inNum, outNum, keyNum;
    int count=0;

Begin:

    end=clock();
    if( (end-begin)/1000 > sec){
        fprintf(fp,"total tested = %d\n", count);
        printf("total tested = %d\n", count);
        return;
    }
    count++;
    inNum=rand() % blockbits; outNum=rand() % blockbits; keyNum=rand() % keybits;
    for(int i=0; i<inNum; i++) inBit[i]=rand() % blockbits;
    for(i=0; i<outNum; i++) outBit[i]=rand() % blockbits;
    for(i=0; i<keyNum; i++) keyBit[i]=rand() % keybits;

    int stat=0;
    for(i=0; i<n1; i++){
        genData(input);
        for(int j=0; j<16; j++) output[j]=input[j];
        genKey(key);
        c->KeyInitial(key, 16, 0);
        c->Encrypt(output, blocksize);
        stat += sepXor128(input, inBit, inNum) ^ 
                sepXor128(output, outBit, outNum) ^ 
                sepXor128(key, keyBit, keyNum);
    }
    double prob;
    prob = ((double) stat) / ((double) n1);
    if(prob > t1 || prob<(1-t1) ){
        for(int i=0; i<n2; i++){
            genData(input);
            for(int j=0; j<16; j++) output[j]=input[j];
            genKey(key);
            c->KeyInitial(key, 16, 0);
            c->Encrypt(output, blocksize);
            stat += sepXor128(input, inBit, inNum) ^ 
                    sepXor128(output, outBit, outNum) ^ 
                    sepXor128(key, keyBit, keyNum);
        }
        prob = ((double) stat) / ((double) (n1+n2));
        if(prob > t2 || prob<(1-t2) ){
            fprintf(fp, "P[");
            for(int k=0; k<inNum-1; k++) fprintf(fp, "%d,", inBit[k]+1);
            fprintf(fp, "%d] xor ", inBit[k]+1);
            fprintf(fp, "C[");
            for(    k=0; k<outNum-1; k++) fprintf(fp, "%d,", outBit[k]+1);
            fprintf(fp, "%d] = ", outBit[k]+1);
            fprintf(fp, "K[");
            for(    k=0; k<keyNum-1; k++) fprintf(fp, "%d,", keyBit[k]+1);
            fprintf(fp, "%d] with probability %f\n", keyBit[k]+1, prob);
            printf("Find One!!\n");
        }
    }
    goto Begin;
}

/*

void main(void){
    CSBase s;
    CBI2MMXBase c;
    linearAtt(&s, 10000, "slinear.txt");
    linearAtt(&c, 10000, "bilinear.txt");
    linearAtt2(&c, 100, 0.6, 1000, 0.58, 7200);
}
*/