#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include "SBase.h"
#include "RC6Base.h"
#include "RC6MMXBase.h"
#include "BI1MMXBase.h"
#include "BI2MMXBase.h"
#include "BI3MMXBase.h"

//unremark the following line if you want to show test output.
//#define show_test_output

//unremart the following line if you want to run test.cpp
#define run_this_test

// n is in Byte
void showBits(BYTE* data, int n){
    for(int i=0; i<n; i++){
        for(int j=0; j<8; j++){
            if(data[i]&( (unsigned int) 0x80 >> j )) printf("1");
            else printf("0");
        }
    }
}

void genKey(BYTE* key){
    int tmp,i;
    for(i=0; i<8; i++){
        tmp=rand();
        mputBWord(key, tmp);
    }
}

void genData(BYTE* data){
    int tmp,i;
    for(i=0; i<8; i++){
        tmp=rand();
        mputBWord(data, tmp);
    }
}

// length in byte
int bitDiff(BYTE* d1, BYTE* d2, int length){
    int result=0,i,j,end=length/4;
    unsigned int tmp1,tmp2, diff, mask;
    for(i=0; i<end; i++){
        tmp1=mgetBLong(d1);
        tmp2=mgetBLong(d2);
        diff=tmp1^tmp2;
        for(j=0,mask=1;j<32; j++,mask<<=1){
            if(diff & mask) result++;
        }
    }
    return result;
}


void timeTest(CCipherBase* Cipher[], double time[], int n){
    clock_t begin, end;
    BYTE key[16];
    BYTE data[16];
    genKey(key);
    genData(data);
    int i,j;
    for(i=0; Cipher[i]!=NULL; i++){
        CCipherBase* C=Cipher[i];
        int blocksize;
        blocksize=C->GetBlockSize();
        C->CipherInitial(0);
        C->KeyInitial(key, 16, 0);
        begin=clock();
        for(j=0; j<n; j++){
            C->Encrypt(data,blocksize);
            C->Decrypt(data,blocksize);
        }
        end=clock();
        time[i]= (end-begin)*1.0/n;
    }
}


void correctTest(CCipherBase* Cipher[], int correct[], int n){
    BYTE key[16];
    BYTE data1[16];
    BYTE data2[16];
    int i,j,k;
    for(i=0; Cipher[i]!=NULL; i++){
        CCipherBase* C=Cipher[i];
        int blocksize;
        blocksize=C->GetBlockSize();
        correct[i]=1;
        for(j=0; j<n; j++){
            genKey(key);
            genData(data1);
            for(k=0; k<16; k++) data2[k]=data1[k];
            C->CipherInitial(0);
            C->KeyInitial(key, 16, 0);
            C->Encrypt(data1,blocksize);
            C->Decrypt(data1,blocksize);
            for(k=0; k<16; k++){
                if(data1[k]!=data2[k]){
                    correct[i]=0;
                    break;
                }
            }
            if(correct[i]==0) break;
        }
    }
}


//plaintext change one bit, how many bits in ciphertext will change.
void Pchange1bitTest(CCipherBase* Cipher[], double mean[], double stdD[], int n){
    BYTE key[16];
    BYTE o_data[16];
    BYTE m_data[16];
    BYTE output[16];
    int i,j,k;
    for(i=0; Cipher[i]!=NULL; i++){
        CCipherBase* C=Cipher[i];
        int blocksize;
        blocksize=C->GetBlockSize();
        mean[i]=0;
        stdD[i]=0;
        for(j=0; j<n; j++){
            int mm=0, std=0;
            genKey(key);
            genData(o_data);
            C->CipherInitial(0);
            C->KeyInitial(key,16,0);
            for(k=0; k<16; k++) output[k]=o_data[k];
            C->Encrypt(output,blocksize);
            for(int m=0; m<blocksize*8; m++){
                for(k=0; k<16; k++) m_data[k]=o_data[k];
                unsigned int mask=0x80;
                mask >>= m%8;
                m_data[m/8]=(BYTE) ((unsigned int) m_data[m/8] ^ mask);
#ifdef show_test_output
                printf("DATA_O:"); showBits(o_data, blocksize); printf("\n");
                printf("DATA_M:"); showBits(m_data, blocksize); printf("\n");
#endif
                C->Encrypt(m_data, blocksize);
                int diff=bitDiff(m_data, output, blocksize);
                mm+=diff;
                std+=diff*diff;
#ifdef show_test_output
                printf("OUTPUT_O:"); showBits(output, blocksize); printf("\n");
                printf("OUTPUT_M:"); showBits(m_data, blocksize); printf("\n");
                printf("difference=%i bits\n", diff);
#endif
            }
            mean[i] += mm*1.0/(blocksize*8.0);
            stdD[i] += std*1.0/(blocksize*8.0);
        }
        mean[i] /= n;
        stdD[i] /= n;
        stdD[i] = sqrt(stdD[i]-(mean[i]*mean[i]));
    }
}


//key change 1 bit, how many bits in ciphertext will change
void Kchange1bitTest(CCipherBase* Cipher[], double mean[], double stdD[], int n){
    BYTE o_key[16];
    BYTE m_key[16];
    BYTE data[16];
    BYTE m_output[16];
    BYTE output[16];
    int i,j,k;
    for(i=0; Cipher[i]!=NULL; i++){
        CCipherBase* C=Cipher[i];
        int blocksize, keysize;
        blocksize=C->GetBlockSize();
        keysize=C->GetKeySize();
        mean[i]=0;
        stdD[i]=0;
        for(j=0; j<n; j++){
            int mm=0, std=0;
            genKey(o_key);
            genData(data);
            C->CipherInitial(0);
            C->KeyInitial(o_key,16,0);
            for(k=0; k<16; k++) output[k]=data[k];
            C->Encrypt(output,blocksize);
            for(int m=0; m<keysize*8; m++){
                for(k=0; k<16; k++) m_output[k]=data[k];
                for(k=0; k<16; k++) m_key[k]=o_key[k];
                unsigned int mask=0x80;
                mask >>= m%8;
                m_key[m/8]=(BYTE) ((unsigned int) m_key[m/8] ^ mask);
#ifdef show_test_output
                printf("KEY_O:"); showBits(o_key, keysize); printf("\n");
                printf("KEY_M:"); showBits(m_key, keysize); printf("\n");
#endif
                C->KeyInitial(m_key,16,0);
                C->Encrypt(m_output, blocksize);
                int diff=bitDiff(m_output, output, blocksize);
                mm+=diff;
                std+=diff*diff;
#ifdef show_test_output
                printf("OUTPUT_O:"); showBits(output, blocksize); printf("\n");
                printf("OUTPUT_M:"); showBits(m_data, blocksize); printf("\n");
                printf("difference=%i bits\n", diff);
#endif
            }
            mean[i] += mm*1.0/(keysize*8.0);
            stdD[i] += std*1.0/(keysize*8.0);
        }
        mean[i] /= n;
        stdD[i] /= n;
        stdD[i] = sqrt(stdD[i]-(mean[i]*mean[i]));
    }
}


//blocksize in byte
void Statistic(BYTE m_data[], BYTE output[], int stat[], int blocksize){
    int i,j,end=blocksize/4;
    unsigned int tmp1,tmp2, diff, mask;
    for(i=0; i<end; i++){
        tmp1=mgetBLong(m_data);
        tmp2=mgetBLong(output);
        diff=tmp1^tmp2;
        for(j=0,mask=1;j<32; j++,mask<<=1){
            if(diff & mask){
                stat[32*i+j]++;
            }
        }
    }
}

//plaintext change a particular bit, what bits will change in ciphertext.
void P1bitStatistic(CCipherBase* Cipher[], double mean[], int n){
    BYTE key[16];
    BYTE o_data[16];
    BYTE m_data[16];
    BYTE output[16];
    static int stat[128][128];
    int i,j,k;
    FILE* fp=fopen("pstat.txt","w");

    for(i=0; Cipher[i]!=NULL; i++){
        CCipherBase* C=Cipher[i];
        int blocksize;
        blocksize=C->GetBlockSize();
        for(int x=0; x<128; x++)
            for(int y=0; y<128; y++) stat[x][y]=0;
        for(j=0; j<n; j++){
            genKey(key);
            genData(o_data);
            C->CipherInitial(0);
            C->KeyInitial(key,16,0);
            for(k=0; k<16; k++) output[k]=o_data[k];
            C->Encrypt(output,blocksize);
            for(int m=0; m<blocksize*8; m++){
                for(k=0; k<16; k++) m_data[k]=o_data[k];
                unsigned int mask=0x80;
                mask >>= m%8;
                m_data[m/8]=(BYTE) ((unsigned int) m_data[m/8] ^ mask);
                C->Encrypt(m_data, blocksize);
                Statistic(m_data, output, stat[m], blocksize);
            }
        }
        double min, max, minMaxDiff=0;
        for(x=0; x<blocksize*8; x++){
            min=max=stat[x][0];
            for(int y=0; y<blocksize*8; y++){
                if(stat[x][y]<min) min=stat[x][y];
                if(stat[x][y]>max) max=stat[x][y];
                fprintf(fp,"%i ", stat[x][y]);
            }
            fprintf(fp, "\n");
            minMaxDiff += (max-min)/max;
        }
        fprintf(fp, "\n");
        mean[i]=minMaxDiff/(blocksize*8.0);
    }
}


//key change a particular bit, what bits will change in cipher text
void K1bitStatistic(CCipherBase* Cipher[], double mean[], int n){
    BYTE m_key[16];
    BYTE o_key[16];
    BYTE o_data[16];
    BYTE t_data[16];
    BYTE output[16];
    static int stat[128][128];
    int i,j,k;
    FILE* fp=fopen("kstat.txt","w");

    for(i=0; Cipher[i]!=NULL; i++){
        CCipherBase* C=Cipher[i];
        int blocksize, keysize;
        blocksize=C->GetBlockSize();
        keysize=C->GetKeySize();
        for(int x=0; x<128; x++)
            for(int y=0; y<128; y++) stat[x][y]=0;
        for(j=0; j<n; j++){
            genKey(o_key);
            genData(o_data);
            C->CipherInitial(0);
            C->KeyInitial(o_key,16,0);
            for(k=0; k<16; k++) output[k]=o_data[k];
            C->Encrypt(output,blocksize);
            for(int m=0; m<keysize*8; m++){
                for(k=0; k<16; k++) m_key[k]=o_key[k];
                unsigned int mask=0x80;
                mask >>= m%8;
                m_key[m/8]=(BYTE) ((unsigned int) m_key[m/8] ^ mask);
                C->KeyInitial(m_key,16,0);
                for(k=0; k<16; k++) t_data[k]=o_data[k];
                C->Encrypt(t_data, blocksize);
                Statistic(t_data, output, stat[m], blocksize);
            }
        }
        double min, max, minMaxDiff=0;
        for(x=0; x<keysize*8; x++){
            min=max=stat[x][0];
            for(int y=0; y<blocksize*8; y++){
                if(stat[x][y]<min) min=stat[x][y];
                if(stat[x][y]>max) max=stat[x][y];
                fprintf(fp,"%i ", stat[x][y]);
            }
            fprintf(fp, "\n");
            minMaxDiff += (max-min)/max;
        }
        fprintf(fp, "\n");
        mean[i]=minMaxDiff/(keysize*8.0);
    }
}


//difference between plaintext and ciphertext
void diffPlainCipher(CCipherBase* Cipher[], double mean[], double stdD[], int n){
    for(int a=0; Cipher[a]!=NULL; a++){
        CCipherBase* C=Cipher[a];
        C->CipherInitial(0);
        BYTE key[16];
        BYTE input[16];
        BYTE output[16];
        int blocksize=C->GetBlockSize();
        mean[a] = 0;
        stdD[a] = 0;
        for(int i=0; i<n; i++){
            genKey(key);
            genData(input);
            for(int k=0; k<16; k++) output[k]=input[k];
            C->KeyInitial(key,16,0);
            C->Encrypt(output, blocksize);
            int diff=bitDiff(input, output, blocksize);
            mean[a] += diff;
            stdD[a] += diff*diff;
        }
        mean[a] /= n;
        stdD[a] /= n;
        stdD[a] = sqrt( stdD[a] - mean[a]*mean[a]);
    }
}


//difference between X xor Y and E(X) xor E(Y)....E means Encrypt
void diffXOR(CCipherBase* Cipher[], double mean[], double stdD[], int n){
    for(int a=0; Cipher[a]!=NULL; a++){
        CCipherBase* C=Cipher[a];
        C->CipherInitial(0);
        BYTE key[16];
        BYTE data1[16];
        BYTE data2[16];
        BYTE inputxor[16];
        BYTE outputxor[16];
        u4byte* D1, *D2, *IN, *OUT;
        D1 = (u4byte*) data1; D2 = (u4byte*) data2;
        IN = (u4byte*) inputxor; OUT = (u4byte*) outputxor;
        int blocksize=C->GetBlockSize();
        mean[a] = 0;
        stdD[a] = 0;
        for(int i=0; i<n; i++){
            genKey(key);
            genData(data1);
            genData(data2);
            for(int k=0; k<4; k++) IN[k] = D1[k] ^ D2[k];
            C->KeyInitial(key,16,0);
            C->Encrypt(data1, blocksize);
            C->Encrypt(data2, blocksize);
            for(k=0; k<4; k++) OUT[k] = D1[k] ^ D2[k];
            int diff=bitDiff(inputxor, outputxor, blocksize);
            mean[a] += diff;
            stdD[a] += diff*diff;
        }
        mean[a] /= n;
        stdD[a] /= n;
        stdD[a] = sqrt( stdD[a] - mean[a]*mean[a]);
    }
}


//difference between E(X xor Y) and E(X) xor E(Y)
void linearXOR(CCipherBase* Cipher[], double mean[], double stdD[], int n){
    for(int a=0; Cipher[a]!=NULL; a++){
        CCipherBase* C=Cipher[a];
        C->CipherInitial(0);
        BYTE key[16];
        BYTE data1[16];
        BYTE data2[16];
        BYTE inputxor[16];
        BYTE outputxor[16];
        u4byte* D1, *D2, *IN, *OUT;
        D1 = (u4byte*) data1; D2 = (u4byte*) data2;
        IN = (u4byte*) inputxor; OUT = (u4byte*) outputxor;
        int blocksize=C->GetBlockSize();
        mean[a] = 0;
        stdD[a] = 0;
        for(int i=0; i<n; i++){
            genKey(key);
            genData(data1);
            genData(data2);
            for(int k=0; k<4; k++) IN[k] = D1[k] ^ D2[k];
            C->KeyInitial(key,16,0);
            C->Encrypt(data1, blocksize);
            C->Encrypt(data2, blocksize);
            C->Encrypt(inputxor, blocksize);
            for(k=0; k<4; k++) OUT[k] = D1[k] ^ D2[k];
            int diff=bitDiff(inputxor, outputxor, blocksize);
            mean[a] += diff;
            stdD[a] += diff*diff;
        }
        mean[a] /= n;
        stdD[a] /= n;
        stdD[a] = sqrt( stdD[a] - mean[a]*mean[a]);
    }
}



#ifdef run_this_test
void main(void){

    CCipherBase* Cipher[4];
    double time[4], Pmean[4], PstdD[4], Kmean[4], KstdD[4], PminMaxDiff[4], KminMaxDiff[4];
    double PCmean[4], PCstdD[4], XORmean[4], XORstdD[4], Lmean[4], LstdD[4];
    int correct[4];

    CSBase s;
    CBI3MMXBase bi3mmx;
    CBI2MMXBase bi2mmx;
    Cipher[0]=&s; Cipher[1]=&bi3mmx; Cipher[2]=NULL;//&bi2mmx; Cipher[3]=NULL;

    char name[3][64]={"S","BI3MMX","BI2MMX"};

    correctTest(Cipher, correct, 1000);
    timeTest(Cipher, time, 1000000);
    Pchange1bitTest(Cipher, Pmean, PstdD, 100);
    Kchange1bitTest(Cipher, Kmean, KstdD, 100);
    P1bitStatistic(Cipher, PminMaxDiff, 1000);
    K1bitStatistic(Cipher, KminMaxDiff, 1000);
    diffPlainCipher(Cipher, PCmean, PCstdD, 10000);
    diffXOR(Cipher, XORmean, XORstdD, 10000);
    linearXOR(Cipher, Lmean, LstdD, 10000);

    printf("***********************************************************************\n");
    printf("correct=1: the implementation is correct, otherwise correct=0\n");
    printf("time: the running time in msec\n");
    printf("P: the number of bits of ciphertext which change value because\n");
    printf("   of plaintext's 1-bit changing\n");
    printf("K: the number of bits of ciphertext which change value because\n");
    printf("   of key's 1-bit changing\n");
    printf("PC: difference in bits between plaintext and cipher text\n");
    printf("XOR: difference between X xor Y and E(X) xor E(Y) where E means Encrypt\n");
    printf("L: difference between E(X xor Y) and E(X) xor E(Y)\n");
    printf("************************************************************************\n\n");

    for(int i=0; i<2; i++){
        printf("%s:\n",name[i]);
        printf("correct=%i time=%f Pmean=%f PstdD=%f\n", correct[i], time[i], Pmean[i], PstdD[i]);
        printf("Kmean=%f KstdD=%f PminMaxDiff=%f KminMaxDiff=%f\n", Kmean[i], KstdD[i], PminMaxDiff[i], KminMaxDiff[i]);
        printf("PCmean=%f PCstdD=%f XORmean=%f XORstdD=%f\n", PCmean[i], PCstdD[i], XORmean[i], XORstdD[i]);
        printf("Lmean=%f LstdD=%f\n\n", Lmean[i], LstdD[i]);
    }
}
#endif